@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #define _GNU_SOURCE
@@ -24,12 +24,11 @@
 #include <stdlib.h>
 #include <string.h>
 #include <math.h>
-#include <limits.h>
 
 #include "jam.h"
 #include "thread.h"
 #include "lock.h"
-#include "arch.h"
+#include "interp.h"
 
 #ifdef DIRECT
 #include "interp-direct.h"
@@ -53,34 +52,34 @@
 
 #define ARRAY_BOUNDS_CHECK(array, idx)                                     \
 {                                                                          \
-    if(idx >= *INST_DATA(array)) {                                         \
+    if(idx >= ARRAY_LEN(array)) {                                          \
         char buff[MAX_INT_DIGITS];                                         \
         snprintf(buff, MAX_INT_DIGITS, "%d", idx);                         \
         THROW_EXCEPTION("java/lang/ArrayIndexOutOfBoundsException", buff); \
     }                                                                      \
 }
 
-u4 *executeJava() {
+uintptr_t *executeJava() {
+    CodePntr pc;
     ExecEnv *ee = getExecEnv();
     Frame *frame = ee->last_frame;
     MethodBlock *mb = frame->mb;
-    u4 *lvars = frame->lvars;
-    u4 *ostack = frame->ostack;
-    CodePntr pc;
+    uintptr_t *lvars = frame->lvars;
+    uintptr_t *ostack = frame->ostack;
     ConstantPool *cp = &(CLASS_CB(mb->class)->constant_pool);
 
     Object *this = (Object*)lvars[0];
     Class *new_class;
     MethodBlock *new_mb;
-    u4 *arg1;
+    uintptr_t *arg1;
 #ifdef PREFETCH
     const void *next_handler;
 #endif
 #ifdef USE_CACHE
     union {
         struct {
-            int v1;
-            int v2;
+            uintptr_t v1;
+            uintptr_t v2;
         } i;
         long long l;
     } cache;
@@ -206,10 +205,10 @@ rewrite_lock:
         PUSH_##level(5, 1);                                \
                                                            \
     DEF_OPC(OPC_FCONST_1, level)                           \
-        PUSH_##level(0x3f800000, 1);                       \
+        PUSH_##level(FLOAT_1_BITS, 1);                     \
                                                            \
     DEF_OPC(OPC_FCONST_2, level)                           \
-        PUSH_##level(0x40000000, 1);                       \
+        PUSH_##level(FLOAT_2_BITS, 1);                     \
                                                            \
     DEF_OPC(OPC_SIPUSH, level)                             \
         PUSH_##level(DOUBLE_SIGNED(pc), 3);                \
@@ -393,13 +392,13 @@ rewrite_lock:
                         "division by zero");
 
 #define ZERO_DIVISOR_CHECK_0                               \
-    ZERO_DIVISOR_CHECK(ostack[-1]);
+    ZERO_DIVISOR_CHECK((int)ostack[-1]);
 
 #define ZERO_DIVISOR_CHECK_1                               \
-    ZERO_DIVISOR_CHECK(cache.i.v1);
+    ZERO_DIVISOR_CHECK((int)cache.i.v1);
 
 #define ZERO_DIVISOR_CHECK_2                               \
-    ZERO_DIVISOR_CHECK(cache.i.v2);
+    ZERO_DIVISOR_CHECK((int)cache.i.v2);
 
 #ifdef USE_CACHE
 #define PUSH_0(value, ins_len)                             \
@@ -492,24 +491,24 @@ rewrite_lock:
 }
 
 #define UNARY_MINUS_1                                      \
-    PUSH_0(-cache.i.v1, 1);
+    PUSH_0(-(int)cache.i.v1, 1);
 
 #define UNARY_MINUS_2                                      \
-    PUSH_1(-cache.i.v2, 1);
+    PUSH_1(-(int)cache.i.v2, 1);
 
 #define BINARY_OP_0(OP)                                    \
     ostack -= 2;                                           \
     PUSH_0((int)ostack[0] OP (int)ostack[1], 1);
 
 #define BINARY_OP_1(OP)                                    \
-    PUSH_0((int)*--ostack OP cache.i.v1, 1);
+    PUSH_0((int)*--ostack OP (int)cache.i.v1, 1);
 
 #define BINARY_OP_2(OP)                                    \
-    PUSH_0(cache.i.v1 OP cache.i.v2, 1);
+    PUSH_0((int)cache.i.v1 OP (int)cache.i.v2, 1);
 
 #define SHIFT_OP_0(TYPE, OP)                               \
     ostack -= 2;                                           \
-    PUSH_0((TYPE)ostack[0] OP ((int)ostack[1] & 0x1f), 1);
+    PUSH_0((TYPE)ostack[0] OP (ostack[1] & 0x1f), 1);
 
 #define SHIFT_OP_1(TYPE, OP)                               \
     PUSH_0((TYPE)*--ostack OP (cache.i.v1 & 0x1f), 1);
@@ -522,24 +521,24 @@ rewrite_lock:
     BRANCH((int)ostack[0] COND (int)ostack[1]);
 
 #define IF_ICMP_1(COND)                                    \
-    BRANCH((int)*--ostack COND cache.i.v1);
+    BRANCH((int)*--ostack COND (int)cache.i.v1);
 
 #define IF_ICMP_2(COND)                                    \
-    BRANCH(cache.i.v1 COND cache.i.v2);
+    BRANCH((int)cache.i.v1 COND (int)cache.i.v2);
 
 #define IF_0(COND)                                         \
     BRANCH((int)*--ostack COND 0);
 
 #define IF_1(COND)                                         \
-    BRANCH(cache.i.v1 COND 0);
+    BRANCH((int)cache.i.v1 COND 0);
 
 #ifdef USE_CACHE
 #define IF_2(COND)                                         \
     *ostack++ = cache.i.v1;                                \
-    BRANCH(cache.i.v2 COND 0);
+    BRANCH((int)cache.i.v2 COND 0);
 #else
 #define IF_2(COND)                                         \
-    BRANCH(cache.i.v2 COND 0);
+    BRANCH((int)cache.i.v2 COND 0);
 #endif
 
 #ifdef DIRECT
@@ -709,7 +708,7 @@ rewrite_lock:
         PUSH_LONG(0, 1);
 
     DEF_OPC_210(OPC_DCONST_1)
-        PUSH_LONG(0x3ff0000000000000LL, 1);
+        PUSH_LONG(DOUBLE_1_BITS, 1);
 
     DEF_OPC_210(OPC_LCONST_1)
         PUSH_LONG(1, 1);
@@ -779,41 +778,31 @@ rewrite_lock:
         POP_LONG(*(u8*)(&lvars[3]), 1);
 
 #ifdef USE_CACHE
-#define ARRAY_LOAD(TYPE)                                 \
-{                                                        \
-    TYPE *element;                                       \
-    int idx = cache.i.v2;                                \
-    Object *array = (Object *)cache.i.v1;                \
-                                                         \
-    NULL_POINTER_CHECK(array);                           \
-    ARRAY_BOUNDS_CHECK(array, idx);                      \
-    element = (TYPE *)(((char *)INST_DATA(array)) +      \
-                              (idx * sizeof(TYPE)) + 4); \
-    cache.i.v1 = *element;                               \
-    DISPATCH(1, 1);                                      \
-}
+#define ARRAY_LOAD_IDX cache.i.v2
+#define ARRAY_LOAD_ARY cache.i.v1
 #else
-#define ARRAY_LOAD(TYPE)                                 \
-{                                                        \
-    TYPE *element;                                       \
-    int idx = *--ostack;                                 \
-    Object *array = (Object *)*--ostack;                 \
-                                                         \
-    NULL_POINTER_CHECK(array);                           \
-    ARRAY_BOUNDS_CHECK(array, idx);                      \
-    element = (TYPE *)(((char *)INST_DATA(array)) +      \
-                              (idx * sizeof(TYPE)) + 4); \
-    *ostack++ = *element;                                \
-    DISPATCH(0, 1);                                      \
-}
+#define ARRAY_LOAD_IDX *--ostack;
+#define ARRAY_LOAD_ARY *--ostack;
 #endif
 
-    DEF_OPC_012_3(
+#define ARRAY_LOAD(TYPE)                              \
+{                                                     \
+    int idx = ARRAY_LOAD_IDX;                         \
+    Object *array = (Object *)ARRAY_LOAD_ARY;         \
+                                                      \
+    NULL_POINTER_CHECK(array);                        \
+    ARRAY_BOUNDS_CHECK(array, idx);                   \
+    PUSH_0(((TYPE *)ARRAY_DATA(array))[idx], 1);      \
+}
+
+    DEF_OPC_012_2(
             OPC_IALOAD,
-            OPC_AALOAD,
             OPC_FALOAD)
         ARRAY_LOAD(int);
 
+    DEF_OPC_012(OPC_AALOAD)
+        ARRAY_LOAD(uintptr_t);
+
     DEF_OPC_012(OPC_BALOAD)
         ARRAY_LOAD(signed char);
 
@@ -827,47 +816,34 @@ rewrite_lock:
             OPC_LALOAD,
             OPC_DALOAD)
     {
-#ifdef USE_CACHE
-        int idx = cache.i.v2;
-        Object *array = (Object *)cache.i.v1;
-#else
-        int idx = *--ostack;
-        Object *array = (Object*)*--ostack;
-#endif
+        int idx = ARRAY_LOAD_IDX;
+        Object *array = (Object *)ARRAY_LOAD_ARY;
 
         NULL_POINTER_CHECK(array);
         ARRAY_BOUNDS_CHECK(array, idx);
-        PUSH_LONG(((u8 *)(INST_DATA(array) + 1))[idx], 1);
+        PUSH_LONG(((u8 *)ARRAY_DATA(array))[idx], 1);
     }
 
 #ifdef USE_CACHE
-#define ARRAY_STORE(TYPE)                               \
-{                                                       \
-    int val = cache.i.v2;                               \
-    int idx = cache.i.v1;                               \
-    Object *array = (Object *)*--ostack;                \
-                                                        \
-    NULL_POINTER_CHECK(array);                          \
-    ARRAY_BOUNDS_CHECK(array, idx);                     \
-    *(TYPE *)(((char *)INST_DATA(array)) +              \
-                       (idx * sizeof(TYPE)) + 4) = val; \
-    DISPATCH(0, 1);                                     \
-}
+#define ARRAY_STORE_VAL cache.i.v2
+#define ARRAY_STORE_IDX cache.i.v1
 #else
-#define ARRAY_STORE(TYPE)                               \
-{                                                       \
-    int val = *--ostack;                                \
-    int idx = *--ostack;                                \
-    Object *array = (Object *)*--ostack;                \
-                                                        \
-    NULL_POINTER_CHECK(array);                          \
-    ARRAY_BOUNDS_CHECK(array, idx);                     \
-    *(TYPE *)(((char *)INST_DATA(array)) +              \
-                       (idx * sizeof(TYPE)) + 4) = val; \
-    DISPATCH(0, 1);                                     \
-}
+#define ARRAY_STORE_VAL *--ostack
+#define ARRAY_STORE_IDX *--ostack
 #endif
 
+#define ARRAY_STORE(TYPE)                         \
+{                                                 \
+    int val = ARRAY_STORE_VAL;                    \
+    int idx = ARRAY_STORE_IDX;                    \
+    Object *array = (Object *)*--ostack;          \
+                                                  \
+    NULL_POINTER_CHECK(array);                    \
+    ARRAY_BOUNDS_CHECK(array, idx);               \
+    ((TYPE *)ARRAY_DATA(array))[idx] = val;       \
+    DISPATCH(0, 1);                               \
+}
+
     DEF_OPC_012_2(
             OPC_IASTORE,
             OPC_FASTORE)
@@ -883,13 +859,8 @@ rewrite_lock:
 
     DEF_OPC_012(OPC_AASTORE)
     { 
-#ifdef USE_CACHE
-        int idx = cache.i.v1;
-        Object *obj = (Object*)cache.i.v2;
-#else
-        Object *obj = (Object*)*--ostack;
-        int idx = *--ostack;
-#endif
+        Object *obj = (Object*)ARRAY_STORE_VAL;
+        int idx = ARRAY_STORE_IDX;
         Object *array = (Object *)*--ostack;
 
         NULL_POINTER_CHECK(array);
@@ -898,7 +869,7 @@ rewrite_lock:
         if((obj != NULL) && !arrayStoreCheck(array->class, obj->class))
             THROW_EXCEPTION("java/lang/ArrayStoreException", NULL);
 
-        INST_DATA(array)[idx + 1] = (u4) obj;
+        ((Object**)ARRAY_DATA(array))[idx] = obj;
         DISPATCH(0, 1);
     }
 
@@ -914,7 +885,7 @@ rewrite_lock:
         NULL_POINTER_CHECK(array);
         ARRAY_BOUNDS_CHECK(array, idx);
 
-        ((u8 *)(INST_DATA(array) + 1))[idx] = cache.l;
+        ((u8 *)ARRAY_DATA(array))[idx] = cache.l;
         DISPATCH(0, 1);
     }
 #else
@@ -929,7 +900,7 @@ rewrite_lock:
         NULL_POINTER_CHECK(array);
         ARRAY_BOUNDS_CHECK(array, idx);
 
-        ((u8 *)(INST_DATA(array) + 1))[idx] = ((u8*)ostack)[1];
+        ((u8 *)ARRAY_DATA(array))[idx] = *(u8*)&ostack[2];
         DISPATCH(0, 1);
     }
 #endif
@@ -945,8 +916,8 @@ rewrite_lock:
         DISPATCH(2, 1);
 
     DEF_OPC_012(OPC_DUP2)
-        *(u8*)ostack = cache.l;
-        ostack += 2;
+        *ostack++ = cache.i.v1;
+        *ostack++ = cache.i.v2;
         DISPATCH(2, 1);
 
     DEF_OPC_012(OPC_DUP2_X1)
@@ -966,7 +937,7 @@ rewrite_lock:
 
     DEF_OPC_012(OPC_SWAP)
     {
-        u4 word1 = cache.i.v1;
+        uintptr_t word1 = cache.i.v1;
         cache.i.v1 = cache.i.v2;
         cache.i.v2 = word1;
         DISPATCH(2, 1);
@@ -974,8 +945,8 @@ rewrite_lock:
 #else /* USE_CACHE */
     DEF_OPC_012(OPC_DUP_X1)
     {
-        u4 word1 = ostack[-1];
-        u4 word2 = ostack[-2];
+        uintptr_t word1 = ostack[-1];
+        uintptr_t word2 = ostack[-2];
         ostack[-2] = word1;
         ostack[-1] = word2;
         *ostack++ = word1;
@@ -984,9 +955,9 @@ rewrite_lock:
 
     DEF_OPC_012(OPC_DUP_X2)
     {
-        u4 word1 = ostack[-1];
-        u4 word2 = ostack[-2];
-        u4 word3 = ostack[-3];
+        uintptr_t word1 = ostack[-1];
+        uintptr_t word2 = ostack[-2];
+        uintptr_t word3 = ostack[-3];
         ostack[-3] = word1;
         ostack[-2] = word3;
         ostack[-1] = word2;
@@ -995,15 +966,16 @@ rewrite_lock:
     }
 
     DEF_OPC_012(OPC_DUP2)
-        *(u8*)ostack = ((u8*)ostack)[-1];
+        ostack[0] = ostack[-2];
+        ostack[1] = ostack[-1];
         ostack += 2;
         DISPATCH(0, 1);
 
     DEF_OPC_012(OPC_DUP2_X1)
     {
-        u4 word1 = ostack[-1];
-        u4 word2 = ostack[-2];
-        u4 word3 = ostack[-3];
+        uintptr_t word1 = ostack[-1];
+        uintptr_t word2 = ostack[-2];
+        uintptr_t word3 = ostack[-3];
         ostack[-3] = word2;
         ostack[-2] = word1;
         ostack[-1] = word3;
@@ -1015,10 +987,10 @@ rewrite_lock:
 
     DEF_OPC_012(OPC_DUP2_X2)
     {
-        u4 word1 = ostack[-1];
-        u4 word2 = ostack[-2];
-        u4 word3 = ostack[-3];
-        u4 word4 = ostack[-4];
+        uintptr_t word1 = ostack[-1];
+        uintptr_t word2 = ostack[-2];
+        uintptr_t word3 = ostack[-3];
+        uintptr_t word4 = ostack[-4];
         ostack[-4] = word2;
         ostack[-3] = word1;
         ostack[-2] = word4;
@@ -1031,16 +1003,18 @@ rewrite_lock:
 
     DEF_OPC_012(OPC_SWAP)
     {
-        u4 word1 = ostack[-1];
+        uintptr_t word1 = ostack[-1];
         ostack[-1] = ostack[-2];
         ostack[-2] = word1;
         DISPATCH(0, 1)
     }
 #endif /* USE_CACHE */
 
-#define BINARY_OP_fp(TYPE, OP)                                        \
-    ((TYPE*)ostack)[-2] = ((TYPE*)ostack)[-2] OP ((TYPE*)ostack)[-1]; \
-    ostack -= sizeof(TYPE)/4;                                         \
+#define BINARY_OP_fp(TYPE, OP)                  \
+    *(TYPE *)&ostack[-sizeof(TYPE)/4 * 2] =     \
+        *(TYPE *)&ostack[-sizeof(TYPE)/4 * 2] OP\
+        *(TYPE *)&ostack[-sizeof(TYPE)/4];      \
+    ostack -= sizeof(TYPE)/4;                   \
     DISPATCH(0, 1);
 
     DEF_OPC_210(OPC_FADD)
@@ -1069,7 +1043,7 @@ rewrite_lock:
 
 #ifdef USE_CACHE
 #define BINARY_OP_long(OP)                               \
-    cache.l = ((long long*)ostack)[-1] OP cache.l;       \
+    cache.l = *(long long*)&ostack[-2] OP cache.l;       \
     ostack -= 2;                                         \
     DISPATCH(2, 1);
 
@@ -1080,7 +1054,7 @@ rewrite_lock:
     BINARY_OP_fp(long long, OP)
 
 #define ZERO_DIVISOR_CHECK_long                          \
-    ZERO_DIVISOR_CHECK(((u8*)ostack)[-1]);
+    ZERO_DIVISOR_CHECK(*(u8*)&ostack[-2]);
 #endif
 
     DEF_OPC_012(OPC_LADD)
@@ -1122,8 +1096,8 @@ rewrite_lock:
 #define SHIFT_OP_long(TYPE, OP)       \
 {                                     \
     int shift = *--ostack & 0x3f;     \
-    ((TYPE*)ostack)[-1] =             \
-        ((TYPE*)ostack)[-1] OP shift; \
+    *(TYPE*)&ostack[-2] =             \
+        *(TYPE*)&ostack[-2] OP shift; \
     DISPATCH(0, 1);                   \
 }
 #endif
@@ -1139,26 +1113,27 @@ rewrite_lock:
 
     DEF_OPC_210(OPC_FREM)
     {
-        float v2 = ((float *)ostack)[-1];
-        float v1 = ((float *)ostack)[-2];
+        float v2 = *(float *)&ostack[-1];
+        float v1 = *(float *)&ostack[-2];
 
-        ((float *)ostack)[-2] = fmod(v1, v2);
+        *(float *)&ostack[-2] = fmod(v1, v2);
         ostack -= 1;
         DISPATCH(0, 1);
     }
 
     DEF_OPC_210(OPC_DREM)
     {
-        double v2 = ((double *)ostack)[-1];
-        double v1 = ((double *)ostack)[-2];
+        double v2 = *(double *)&ostack[-2];
+        double v1 = *(double *)&ostack[-4];
 
-        ((double *)ostack)[-2] = fmod(v1, v2);
+        *(double *)&ostack[-4] = fmod(v1, v2);
         ostack -= 2;
         DISPATCH(0, 1);
     }
 
-#define UNARY_MINUS(TYPE)                       \
-    ((TYPE*)ostack)[-1] = -((TYPE*)ostack)[-1]; \
+#define UNARY_MINUS(TYPE)                    \
+    *(TYPE*)&ostack[-sizeof(TYPE)/4] =       \
+          -*(TYPE*)&ostack[-sizeof(TYPE)/4]; \
     DISPATCH(0, 1);
 
     DEF_OPC_210(OPC_LNEG)
@@ -1170,6 +1145,35 @@ rewrite_lock:
     DEF_OPC_210(OPC_DNEG)
         UNARY_MINUS(double);
 
+    DEF_OPC_210(OPC_I2L)
+        ostack -= 1;
+        PUSH_LONG((int)*ostack, 1);
+
+    DEF_OPC_012(OPC_L2I)
+    {
+       long long l;
+#ifdef USE_CACHE
+        l = cache.l;
+#else
+        ostack -= 2;
+        l = *(long long*)ostack;
+#endif
+        PUSH_0((int)l, 1);
+    }
+
+#define OPC_int2fp(DEST_TYPE)            \
+    ostack -= 1;                         \
+    *(DEST_TYPE *)ostack =               \
+              (DEST_TYPE)(int)*ostack;   \
+    ostack += sizeof(DEST_TYPE)/4;       \
+    DISPATCH(0, 1);
+
+    DEF_OPC_210(OPC_I2F)
+        OPC_int2fp(float);
+
+    DEF_OPC_210(OPC_I2D)
+        OPC_int2fp(double);
+
 #define OPC_X2Y(SRC_TYPE, DEST_TYPE)     \
 {                                        \
     SRC_TYPE v;                          \
@@ -1180,18 +1184,6 @@ rewrite_lock:
     DISPATCH(0, 1);                      \
 }
 
-    DEF_OPC_210(OPC_I2L)
-        OPC_X2Y(int, long long);
-
-    DEF_OPC_210(OPC_I2F)
-        OPC_X2Y(int, float);
-
-    DEF_OPC_210(OPC_I2D)
-        OPC_X2Y(int, double);
-
-    DEF_OPC_210(OPC_L2I)
-        OPC_X2Y(long long, int);
-
     DEF_OPC_210(OPC_L2F)
         OPC_X2Y(long long, float);
 
@@ -1229,23 +1221,23 @@ rewrite_lock:
     DEF_OPC_210(OPC_D2I)
         OPC_fp2int(double);
 
-#define OPC_fp2long(SRC_TYPE)                  \
-{                                              \
-    long long res;                             \
-    SRC_TYPE value;                            \
-    ostack -= sizeof(SRC_TYPE)/4;              \
-    value = *(SRC_TYPE *)ostack;               \
-                                               \
-    if(value >= (SRC_TYPE)LONG_LONG_MAX)       \
-        res = LONG_LONG_MAX;                   \
-    else if(value <= (SRC_TYPE)LONG_LONG_MIN)  \
-        res = LONG_LONG_MIN;                   \
-    else if(value != value)                    \
-        res = 0;                               \
-    else                                       \
-        res = (long long) value;               \
-                                               \
-    PUSH_LONG(res, 1);                         \
+#define OPC_fp2long(SRC_TYPE)              \
+{                                          \
+    long long res;                         \
+    SRC_TYPE value;                        \
+    ostack -= sizeof(SRC_TYPE)/4;          \
+    value = *(SRC_TYPE *)ostack;           \
+                                           \
+    if(value >= (SRC_TYPE)LLONG_MAX)       \
+        res = LLONG_MAX;                   \
+    else if(value <= (SRC_TYPE)LLONG_MIN)  \
+        res = LLONG_MIN;                   \
+    else if(value != value)                \
+        res = 0;                           \
+    else                                   \
+        res = (long long) value;           \
+                                           \
+    PUSH_LONG(res, 1);                     \
 }
 
     DEF_OPC_210(OPC_F2L)
@@ -1275,7 +1267,7 @@ rewrite_lock:
 #ifdef USE_CACHE
     DEF_OPC_012(OPC_LCMP)
     {
-        long long v1 = ((long long*)ostack)[-1];
+        long long v1 = *(long long*)&ostack[-2];
         int r = (v1 == cache.l) ? 0 : ((v1 < cache.l) ? -1 : 1);
         cache.i.v1 = r;
         ostack -= 2;
@@ -1284,8 +1276,8 @@ rewrite_lock:
 #else
     DEF_OPC_012(OPC_LCMP)
     {
-        long long v2 = ((long long*)ostack)[-1];
-        long long v1 = ((long long*)ostack)[-2];
+        long long v2 = *(long long*)&ostack[-2];
+        long long v1 = *(long long*)&ostack[-4];
         ostack[-4] = (v1 == v2) ? 0 : ((v1 < v2) ? -1 : 1);
         ostack -= 3;
         DISPATCH(0, 1);
@@ -1337,7 +1329,7 @@ rewrite_lock:
 #else
     DEF_OPC_210(OPC_JSR)
 #endif
-        *ostack++ = (u4)pc;
+        *ostack++ = (uintptr_t)pc;
         BRANCH(TRUE);
 
     DEF_OPC_210(OPC_RET)
@@ -1361,7 +1353,7 @@ rewrite_lock:
         Object *array = (Object *)*--ostack;
 
         NULL_POINTER_CHECK(array);
-        PUSH_0(*INST_DATA(array), 1);
+        PUSH_0(ARRAY_LEN(array), 1);
     }
 
     DEF_OPC_210(OPC_ATHROW)
@@ -1384,7 +1376,7 @@ rewrite_lock:
         if((obj = allocTypeArray(type, count)) == NULL)
             goto throwException;
 
-        PUSH_0((u4)obj, 2);
+        PUSH_0((uintptr_t)obj, 2);
     }
 
     DEF_OPC_210(OPC_MONITORENTER)
@@ -1657,7 +1649,7 @@ rewrite_lock:
     }
 #else /* DIRECT */
     DEF_OPC_210(OPC_LDC)
-        frame->last_pc = (unsigned char*)pc;
+        frame->last_pc = pc;
 
         resolveSingleConstant(mb->class, SINGLE_INDEX(pc));
 
@@ -1668,7 +1660,7 @@ rewrite_lock:
         DISPATCH(0, 0);
 
     DEF_OPC_210(OPC_LDC_W)
-        frame->last_pc = (unsigned char*)pc;
+        frame->last_pc = pc;
 
         resolveSingleConstant(mb->class, DOUBLE_INDEX(pc));
 
@@ -1687,7 +1679,7 @@ rewrite_lock:
 
     DEF_OPC_210(OPC_TABLESWITCH)
     {
-        int *aligned_pc = (int*)((int)(pc + 4) & ~0x3);
+        int *aligned_pc = (int*)((uintptr_t)(pc + 4) & ~0x3);
         int deflt = ntohl(aligned_pc[0]);
         int low   = ntohl(aligned_pc[1]);
         int high  = ntohl(aligned_pc[2]);
@@ -1699,7 +1691,7 @@ rewrite_lock:
 
     DEF_OPC_210(OPC_LOOKUPSWITCH)
     {
-        int *aligned_pc = (int*)((int)(pc + 4) & ~0x3);
+        int *aligned_pc = (int*)((uintptr_t)(pc + 4) & ~0x3);
         int deflt  = ntohl(aligned_pc[0]);
         int npairs = ntohl(aligned_pc[1]);
         int key    = *--ostack;
@@ -1714,7 +1706,7 @@ rewrite_lock:
     {
         FieldBlock *fb;
                
-        frame->last_pc = (unsigned char*)pc;
+        frame->last_pc = pc;
         fb = resolveField(mb->class, DOUBLE_INDEX(pc));
 
         if(exceptionOccured0(ee))
@@ -1731,7 +1723,7 @@ rewrite_lock:
     {
         FieldBlock *fb;
                
-        frame->last_pc = (unsigned char*)pc;
+        frame->last_pc = pc;
         fb = resolveField(mb->class, DOUBLE_INDEX(pc));
 
         if(exceptionOccured0(ee))
@@ -1751,7 +1743,7 @@ rewrite_lock:
 
         WITH_OPCODE_CHANGE_CP_DINDEX(OPC_GETFIELD, idx);
 
-        frame->last_pc = (unsigned char*)pc;
+        frame->last_pc = pc;
         fb = resolveField(mb->class, idx);
 
         if(exceptionOccured0(ee))
@@ -1773,7 +1765,7 @@ rewrite_lock:
 
         WITH_OPCODE_CHANGE_CP_DINDEX(OPC_PUTFIELD, idx);
 
-        frame->last_pc = (unsigned char*)pc;
+        frame->last_pc = pc;
         fb = resolveField(mb->class, idx);
 
         if(exceptionOccured0(ee))
@@ -1792,7 +1784,7 @@ rewrite_lock:
     {
         FieldBlock *fb = RESOLVED_FIELD(pc);
         Object *obj = (Object *)*--ostack;
-        u4 *addr;
+        uintptr_t *addr;
 
         NULL_POINTER_CHECK(obj);
         addr = &(INST_DATA(obj)[fb->offset]);
@@ -1849,7 +1841,7 @@ rewrite_lock:
         int idx;
         WITH_OPCODE_CHANGE_CP_DINDEX(OPC_INVOKEVIRTUAL, idx);
 
-        frame->last_pc = (unsigned char*)pc;
+        frame->last_pc = pc;
         new_mb = resolveMethod(mb->class, idx);
  
         if(exceptionOccured0(ee))
@@ -1878,7 +1870,7 @@ rewrite_lock:
         int idx;
         WITH_OPCODE_CHANGE_CP_DINDEX(OPC_INVOKESPECIAL, idx);
 
-        frame->last_pc = (unsigned char*)pc;
+        frame->last_pc = pc;
         new_mb = resolveMethod(mb->class, idx);
  
         if(exceptionOccured0(ee))
@@ -1896,7 +1888,7 @@ rewrite_lock:
     }
 
     DEF_OPC_210(OPC_INVOKESTATIC)
-        frame->last_pc = (unsigned char*)pc;
+        frame->last_pc = pc;
         new_mb = resolveMethod(mb->class, DOUBLE_INDEX(pc));
  
         if(exceptionOccured0(ee))
@@ -1906,7 +1898,7 @@ rewrite_lock:
         DISPATCH(0, 0);
 
     DEF_OPC_210(OPC_INVOKEINTERFACE)
-        frame->last_pc = (unsigned char*)pc;
+        frame->last_pc = pc;
         new_mb = resolveInterfaceMethod(mb->class, DOUBLE_INDEX(pc));
  
         if(exceptionOccured0(ee))
@@ -1927,7 +1919,7 @@ rewrite_lock:
             OPC_CHECKCAST,
             OPC_INSTANCEOF,
             OPC_MULTIANEWARRAY)
-        frame->last_pc = (unsigned char*)pc;
+        frame->last_pc = pc;
         resolveClass(mb->class, DOUBLE_INDEX(pc), *pc == OPC_NEW);
 
         if(exceptionOccured0(ee))
@@ -1984,7 +1976,7 @@ rewrite_lock:
         DISPATCH(0, READ_S4_OP(pc));
 
     DEF_OPC_210(OPC_JSR_W)
-        PUSH_0((u4)pc+2, READ_S4_OP(pc));
+        PUSH_0((uintptr_t)pc+2, READ_S4_OP(pc));
 
     DEF_OPC_210(OPC_LOCK)
         DISPATCH(0, 0);
@@ -2110,7 +2102,7 @@ rewrite_lock:
         if((obj = allocObject(class)) == NULL)
             goto throwException;
 
-        PUSH_0((u4)obj, 3);
+        PUSH_0((uintptr_t)obj, 3);
     }
  
     DEF_OPC_210(OPC_ANEWARRAY_QUICK)
@@ -2136,16 +2128,16 @@ rewrite_lock:
         else
             strcat(strcat(strcpy(ac_name, "[L"), name), ";");
 
-        array_class = findArrayClassFromClass(ac_name, class);
+        array_class = findArrayClassFromClass(ac_name, mb->class);
         free(ac_name);
 
         if(exceptionOccured0(ee))
             goto throwException;
 
-        if((obj = allocArray(array_class, count, 4)) == NULL)
+        if((obj = allocArray(array_class, count, sizeof(Object*))) == NULL)
             goto throwException;
 
-        PUSH_0((u4)obj, 3);
+        PUSH_0((uintptr_t)obj, 3);
     }
 
     DEF_OPC_210(OPC_CHECKCAST_QUICK)
@@ -2180,15 +2172,15 @@ rewrite_lock:
         frame->last_pc = pc;
 
         for(i = 0; i < dim; i++)
-            if((int)ostack[i] < 0) {
+            if((intptr_t)ostack[i] < 0) {
                 signalException("java/lang/NegativeArraySizeException", NULL);
                 goto throwException;
             }
 
-        if((obj = allocMultiArray(class, dim, ostack)) == NULL)
+        if((obj = allocMultiArray(class, dim, (intptr_t *)ostack)) == NULL)
             goto throwException;
 
-        PUSH_0((u4)obj, 4);
+        PUSH_0((uintptr_t)obj, 4);
     }
 
     DEF_OPC_210(OPC_INVOKEVIRTUAL_QUICK)
@@ -2207,7 +2199,7 @@ invokeMethod:
     Object *sync_ob = NULL;
 
     frame->last_pc = pc;
-    ostack = (u4*)(new_frame+1);
+    ostack = (uintptr_t *)(new_frame+1);
 
     if((char*)(ostack + new_mb->max_stack) > ee->stack_end) {
         if(ee->overflow++) {
@@ -2234,7 +2226,7 @@ invokeMethod:
     }
 
     if(new_mb->access_flags & ACC_NATIVE) {
-        ostack = (*(u4 *(*)(Class*, MethodBlock*, u4*))
+        ostack = (*(uintptr_t *(*)(Class*, MethodBlock*, uintptr_t*))
                      new_mb->native_invoker)(new_mb->class, new_mb, arg1);
 
         if(sync_ob)
@@ -2320,7 +2312,7 @@ throwException:
         this = (Object*)lvars[0];
         cp = &(CLASS_CB(mb->class)->constant_pool);
 
-        *ostack++ = (u4)excep;
+        *ostack++ = (uintptr_t)excep;
 
         /* Dispatch to the first bytecode */
 
