@@ -15,9 +15,36 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
+#include <limits.h>
+
+/* _GNU_SOURCE doesn't bring in C99 long long constants,
+   but we do get the GNU constants */
+#ifndef LLONG_MAX
+#define LLONG_MAX LONG_LONG_MAX
+#endif
+
+#ifndef LLONG_MIN
+#define LLONG_MIN LONG_LONG_MIN
+#endif
+
+#ifdef HAVE_ENDIAN_H
+#include <endian.h>
+#elif HAVE_SYS_PARAM_H
+#include <sys/param.h>
+#endif
+
+#if BYTE_ORDER == BIG_ENDIAN
+#define FLOAT_CONST(val) (uintptr_t)val << (sizeof(uintptr_t) * 8 - 32)
+#else
+#define FLOAT_CONST(val) val
+#endif
+
+#define FLOAT_1_BITS FLOAT_CONST(0x3f800000)
+#define FLOAT_2_BITS FLOAT_CONST(0x40000000)
+
 #define READ_U1_OP(p)    ((p)[1])
 #define READ_U2_OP(p)    (((p)[1]<<8)|(p)[2])
 #define READ_U4_OP(p)    (((p)[1]<<24)|((p)[2]<<16)|((p)[3]<<8)|(p)[4])
