@@ -15,9 +15,10 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
+#ifndef USE_FFI
 #include <stdio.h>
 #include "../../../jam.h"
 
@@ -46,3 +47,4 @@ int nativeExtraArg(MethodBlock *mb) {
     return 4;
 #endif
 }
+#endif
