@@ -15,13 +15,12 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <stdio.h>
 #include <string.h>
 #include "jam.h"
-#include "arch.h"
 
 MethodBlock *findMethod(Class *class, char *methodname, char *type) {
    ClassBlock *cb = CLASS_CB(class);
@@ -118,9 +117,14 @@ retry:
             if(resolved_class == NULL)
                 return NULL;
 
+            if(!checkClassAccess(resolved_class, class)) {
+                signalException("java/lang/IllegalAccessException", "class is not accessible");
+                return NULL;
+            }
+
             CP_TYPE(cp, cp_index) = CONSTANT_Locked;
             MBARRIER();
-            CP_INFO(cp, cp_index) = (u4)resolved_class;
+            CP_INFO(cp, cp_index) = (uintptr_t)resolved_class;
             MBARRIER();
             CP_TYPE(cp, cp_index) = CONSTANT_Resolved;
 
@@ -179,11 +183,16 @@ retry:
                     return NULL;
                 }
 
+                if(!checkMethodAccess(mb, class)) {
+                    signalException("java/lang/IllegalAccessException", "method is not accessible");
+                    return NULL;
+                }
+
                 initClass(mb->class);
 
                 CP_TYPE(cp, cp_index) = CONSTANT_Locked;
                 MBARRIER();
-                CP_INFO(cp, cp_index) = (u4)mb;
+                CP_INFO(cp, cp_index) = (uintptr_t)mb;
                 MBARRIER();
                 CP_TYPE(cp, cp_index) = CONSTANT_Resolved;
             } else
@@ -243,7 +252,7 @@ retry:
             if(mb) {
                 CP_TYPE(cp, cp_index) = CONSTANT_Locked;
                 MBARRIER();
-                CP_INFO(cp, cp_index) = (u4)mb;
+                CP_INFO(cp, cp_index) = (uintptr_t)mb;
                 MBARRIER();
                 CP_TYPE(cp, cp_index) = CONSTANT_Resolved;
             } else
@@ -288,11 +297,16 @@ retry:
             fb = lookupField(resolved_class, fieldname, fieldtype);
 
             if(fb) {
+                if(!checkFieldAccess(fb, class)) {
+                    signalException("java/lang/IllegalAccessException", "field is not accessible");
+                    return NULL;
+                }
+
                 initClass(fb->class);
 
                 CP_TYPE(cp, cp_index) = CONSTANT_Locked;
                 MBARRIER();
-                CP_INFO(cp, cp_index) = (u4)fb;
+                CP_INFO(cp, cp_index) = (uintptr_t)fb;
                 MBARRIER();
                 CP_TYPE(cp, cp_index) = CONSTANT_Resolved;
             } else
@@ -305,7 +319,7 @@ retry:
     return fb;
 }
 
-u4 resolveSingleConstant(Class *class, int cp_index) {
+uintptr_t resolveSingleConstant(Class *class, int cp_index) {
     ConstantPool *cp = &(CLASS_CB(class)->constant_pool);
 
 retry:
@@ -329,7 +343,7 @@ retry:
             if(string) {
                 CP_TYPE(cp, cp_index) = CONSTANT_Locked;
                 MBARRIER();
-                CP_INFO(cp, cp_index) = (u4)findInternedString(string);
+                CP_INFO(cp, cp_index) = (uintptr_t)findInternedString(string);
                 MBARRIER();
                 CP_TYPE(cp, cp_index) = CONSTANT_Resolved;
             }
